#!/usr/bin/env python

# Screenlets main module (c) RYX (aka Rico Pfaus) 2007 <ryx@ryxperience.com>
#
##@mainpage
#
##@section intro_sec General Information
#
# INFO:
# - Screenlets are small owner-drawn applications that can be described as
#  " the virtual representation of things lying/standing around on your desk".
#   Sticknotes, clocks, rulers, ... the possibilities are endless. The goal of the 
#   Screenlets is to simplify the creation of fully themeable mini-apps that
#   each solve basic desktop-work-related needs and generally improve the 
#   usability and eye-candy of the modern Linux-desktop.
#
# TODO: 
# - still more error-handling !!!
# - improve xml-based menu (is implemented, but I'm not happy with it)
# - switching themes slowly increases the memory usage (possible leak)
# - maybe attributes for dependancies/requirements (e.g. special 
#   python-libs or certain Screenlets)
# - ScreenletTheme needs support for overriding handlers in Screenlet
# -
#

import pygtk
pygtk.require('2.0')
import gtk
import math
import cairo
from datetime import datetime
import gobject
import rsvg
import sys
import os
import glob
import dbus

from options import *

# TEST
import XmlMenu
# /TEST


#-------------------------------------------------------------------------------
# CONSTANTS
#-------------------------------------------------------------------------------

# the version of the Screenlets-baseclass in use
VERSION = "0.0.7"

# the install prefix (/usr or /usr/local) DO NOT CHANGE YET!!!
INSTALL_PREFIX = '/usr'

# the global PATH where the screenlets are installed 
PATH = INSTALL_PREFIX + '/share/screenlets'

# dbus-access to backend
BACKEND_BUS = 'org.freedesktop.Screenlets'
BACKEND_OBJ = '/org/freedesktop/Screenlets'
BACKEND_IFACE = 'org.freedesktop.Screenlets'

# A list containing all the paths to search for screenlet-"packages"
# (these paths get searched when a new screenlet-instance shall be
# loaded through the module-loader function)
SCREENLETS_PATH = [os.environ['HOME'] + '/.screenlets',
	'/usr/share/screenlets', '/usr/local/share/screenlets', ]



#-------------------------------------------------------------------------------
# CLASSES
#-------------------------------------------------------------------------------

class DefaultMenuItem:
	"""A container with constants for the default menuitems"""
	
	# default menuitem constants (is it right to increase like this?)
	NONE = 0
	DELETE = 1
	THEMES = 2
	INFO = 4
	SIZE = 8
	WINDOW_MENU = 16
	PROPERTIES = 32
	# EXPERIMENTAL!! If you use this, the file menu.xml in the 
	# Screenlet's data-dir is used for generating the menu ...
	XML = 512
	# the default items
	STANDARD  = 1|2|8|16|32


class ScreenletTheme(dict):
	"""ScreenletThemes are simple storages that allow loading files
as svg-handles within a theme-directory. Each Screenlet can have 
its own theme-directory. It is up to the Screenlet-developer if he
wants to let his Screenlet support themes or not. Themes are 
turned off by default - if your Screenlet uses Themes, just set the 
attribute 'theme_name' to the name of the theme's dir you want to use."""
	
	# meta-info (set through theme.conf)
	__name__ = ''
	__author__ = ''
	__version__ = ''
	__info__ = ''
	
	# attributes
	path = ""
	loaded = False
	width = 0
	height = 0
	option_overrides = {}
	
	def __init__(self, path):
		# set theme-path and load all files in path
		self.path = path
		self.option_overrides = {}
		self.loaded = self.__load_all()
		if self.loaded == False:
			print "Error while loading ScreenletTheme in: " + path
	
	def __getattr__(self, name):
		if name in ("width", "height"):
			if self.loaded and len(self)>0:
				size=self[0].get_dimension_data()
				if name=="width":
					return size[0]
				else:
					return size[1]
		else:
			return object.__getattr__(self, name)
	
	def apply_option_overrides(self, screenlet):
		"""Apply this theme's overridden options to the given Screenlet."""
		for name in self.option_overrides:
			print "Override: "+name
			if self.option_overrides.has_key(name):
				o = screenlet.get_option_by_name(name)
				if o:
					# set option in screenlet
					setattr(screenlet, name, 
						o.on_import(self.option_overrides[name]))
		
	def check_entry(self, filename):
		"""Checks if a file with filename is loaded in this theme."""
		try:
			if self[filename]:
				return True
		except:
			#raise Exception
			return False
	
	def has_overrides(self):
		"""Check if this theme contains overrides for options."""
		if len(self.option_overrides) > 0:
			return True
		return False
	
	def load_conf(self, filename):
		"""Load a config-file from this theme's dir and save vars in list."""
		f=None
		try:
			f = open (filename, "r")
		except:
			screenlets.show_error(self, "file " + str(filename) + " not found")
		if f:
			lines = f.readlines()
			for line in lines:
				# skip comment, section, empty and EOL 
				if line[0] in ("[", "#", " ", "\n"):
					pass
				elif line[0] == "_":
					# underscore? check for theme's meta-info
					if line.startswith('_name='):
						self.__name__ = line[6:][:-1]
					elif line.startswith('_author='):
						self.__author__ = line[8:][:-1]
					elif line.startswith('_version='):
						self.__version__ = line[9:][:-1]
					elif line.startswith('_info='):
						self.__info__ = line[6:][:-1]
				else:
					# read option and add it to list
					tmp = line.split('=', 1)
					if len(tmp) > 0:
						var = tmp[0]
						val = tmp[1][:-1]
						print var + "=" + val.replace("\n", "")
						self.option_overrides[var] = val
					else:
						print "ScreenletTheme: ignored illegal line in " +\
							filename+": '"+line+"'"
			f.close()
			# TEST
			print "theme.conf loaded: "
			print "Name: " + self.__name__
			print "Author: " +self.__author__
			print "Version: " +self.__version__
			print "Info: " +self.__info__
	
	def load_svg(self, filename):
		"""Load an SVG-file into this theme and reference it as ref_name."""
		self[filename] = rsvg.Handle(self.path + "/" + filename)
		if self[filename] != None:
			# set width/height
			size=self[filename].get_dimension_data()
			if size:
				self.width = size[0]
				self.height = size[1]
			return True
		else:
			return False
	
	def load_png(self, filename):
		"""Load a PNG-file into this theme and reference it as ref_name."""
		self[filename] = cairo.ImageSurface.create_from_png(self.path + 
			"/" + filename)
		if self[filename] != None:
			return True
		else:
			return False
	
	def __load_all(self):
		"""Load all files in the theme's path. Currently only loads SVGs."""
		# clear overrides
		#self.__option_overrides = {}
		# read dir
		dirlst = glob.glob(self.path + '/*')
		if len(dirlst)==0:
			return False
		plen = len(self.path) + 1
		for file in dirlst:
			fname = file[plen:]
			if fname.endswith('.svg'):
				# svg file
				if self.load_svg(fname) == False:
					return False
			elif fname.endswith('.png'):
				# svg file
				if self.load_png(fname) == False:
					return False
			elif fname == "theme.conf":
				print "theme.conf found! Loading option-overrides."
				# theme.conf
				if self.load_conf(file) == False:
					return False
		return True
	
	def reload(self):
		"""Re-Load all files in the theme's path."""
		self.free()
		self.__load_all()
	
	# TODO: fix function, rsvg handles are not freed properly
	def free(self):
		"""Deletes the Theme's contents and frees all rsvg-handles."""
		self.option_overrides.clear()
		for filename in self:
			#del self[filename].close()
			del filename
			pass


class Screenlet(gobject.GObject, EditableOptions):
	"""A Screenlet is a (i.e. contains a) shaped gtk-window that is
fully invisible by default. Subclasses of Screenlet can render 
their owner-drawn graphics on fully transparent background."""
	
	# default meta-info for Screenlets
	__name__ ='No name set for this Screenlet'
	__version__ = '0.0'
	__author__ = 'No author defined for this Screenlet'
	__desc__ = 'No info set for this Screenlet'
	__requires__ = ['dbus', 'xdg', 'rsvg']
	#__target_version__ = '0.0.0'
	#__backend_version__ = '0.0.1'
	
	# these are set internally !!!!!
	__path__ = ''	# the screenlets path (os.getcwd() or set by backend)
	__id = None		# internal id (workaround: needed for dbus/settings)
	
	# notifications/dbus support
	__NOTIFY_BUS = 'org.freedesktop.Notifications'
	__NOTIFY_OBJ = '/org/freedesktop/Notifications'
	__NOTIFY_IFACE =  'org.freedesktop.Notifications'
	bus = dbus.SessionBus()
	notifications = dbus.Interface(bus.get_object(__NOTIFY_BUS, 
		__NOTIFY_OBJ), __NOTIFY_IFACE)
	
	# attributes
	window = None
	theme = None
	menu = None
	is_dragged = False
	quit_on_close = True
	allow_option_override = True
	
	# editable options
	x = 0
	y = 0
	width = 100
	height = 100
	scale = 1.0
	theme_name = ""
	is_sticky = False
	is_widget = False
	keep_above = True
	keep_below = False
	skip_pager = True
	skip_taskbar = True

	# some menuitems (needed for checking/unchecking)
	# DEPRECATED: remove - don't really work anyway ...
	__mi_keep_above = None
	__mi_keep_below = None
	__mi_widget = None
	__mi_sticky = None
	
	# for custom signals
	__gsignals__ = dict(screenlet_removed=(gobject.SIGNAL_RUN_FIRST,
		gobject.TYPE_NONE, (gobject.TYPE_OBJECT,)))

	def __init__(self, width=100, height=100, parent_window=None, 
		show_window=True, is_widget=False, is_sticky=False, 
		uses_theme=False, path=os.getcwd()):
		"""Constructor - should only be subclassed"""
		# call gobject and EditableOptions superclasses
		super(Screenlet, self).__init__()
		EditableOptions.__init__(self)
		# init properties
		self.width = width
		self.height = height
		self.is_dragged = False
		self.__path__ = path
		# set some attributes without calling __setattr__
		self.__dict__['theme_name'] = ""
		self.__dict__['is_widget'] = is_widget
		self.__dict__['is_sticky'] = is_sticky
		self.__dict__['x'] = 0
		self.__dict__['y'] = 0
		# TEST: set scale relative to theme size (NOT WORKING)
		#self.__dict__['scale'] = width/100.0
		# /TEST
		# TEST: dbus-based settings-backend
		self.backend = dbus.Interface(self.bus.get_object(BACKEND_BUS, 
			BACKEND_OBJ), BACKEND_IFACE)
		# "editable" options
		self.add_options_group('Screenlet', 
			'The basic settings for this Screenlet-instance.')
		if uses_theme:
			self.add_option(StringOption('Screenlet', 'theme_name', 
				'default', 'Theme', 
				'The theme (design/style) for this Screenlet ...',
				choices = self.get_available_themes()))
		self.add_option(IntOption('Screenlet', 'x', 
			0, 'X-Position', 'The X-position of this Screenlet ...', 
			min=0, max=gtk.gdk.screen_width()))
		self.add_option(IntOption('Screenlet', 'y', 
			0, 'Y-Position', 'The Y-position of this Screenlet ...', 
			min=0, max=gtk.gdk.screen_height()))
		self.add_option(IntOption('Screenlet', 'width', 
			width, 'Width', 'The width of this Screenlet ...', 
			min=16, max=1000))
		self.add_option(IntOption('Screenlet', 'height', 
			height, 'Height', 'The height of this Screenlet ...', 
			min=16, max=1000))
		self.add_option(FloatOption('Screenlet', 'scale', 
			self.scale, 'Scale', 'The scale-factor of this Screenlet ...', 
			min=0.1, max=10.0, digits=2))
		self.add_option(BoolOption('Screenlet', 'is_sticky', 
			is_sticky, 'Stick to Desktop', 
			'Show this Screenlet on all workspaces ...'))
		self.add_option(BoolOption('Screenlet', 'is_widget', 
			is_widget, 'Treat as Widget', 
			'Treat this Screenlet as a "Widget" ...'))
		self.add_option(BoolOption('Screenlet', 'keep_above', 
			self.keep_above, 'Keep above', 
			'Keep this Screenlet above other windows ...'))
		self.add_option(BoolOption('Screenlet', 'keep_below', 
			self.keep_below, 'Keep below', 
			'Keep this Screenlet below other windows ...'))
		self.add_option(BoolOption('Screenlet', 'skip_pager', 
			self.skip_pager, 'Skip Pager', 
			'Set this Screenlet to show/hide in pagers ...'))
		self.add_option(BoolOption('Screenlet', 'skip_taskbar', 
			self.skip_pager, 'Skip Taskbar', 
			'Set this Screenlet to show/hide in taskbars ...'))
		# disable width/height
		self.disable_option('width')
		self.disable_option('height')
		# create window
		self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
		if parent_window:
			self.window.set_parent_window(parent_window)
			self.window.set_transient_for(parent_window)
			self.window.set_destroy_with_parent(True)
		self.window.resize(width, height)
		self.window.set_decorated(False)
		self.window.set_app_paintable(True)
		# set type hint
		#self.window.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_DOCK)
		self.window.set_keep_above(True)
		self.window.set_skip_taskbar_hint(True)
		self.window.set_skip_pager_hint(True)
		if is_sticky:
			self.window.stick()
		self.alpha_screen_changed(self.window)
		self.update_shape()
		#self.window.set_events(gtk.gdk.BUTTON_PRESS_MASK)
		self.window.set_events(gtk.gdk.ALL_EVENTS_MASK)
		self.window.connect("delete_event", self.delete_event)
		self.window.connect("destroy", self.destroy)
		self.window.connect("expose_event", self.expose)
		self.window.connect("button-press-event", self.button_press)
		self.window.connect("button-release-event", self.button_release)
		self.window.connect("drag-begin", self.drag_begin)
		self.window.connect("drag-leave", self.drag_end)
		self.window.connect("configure-event", self.configure_event)
		self.window.connect("screen-changed", self.alpha_screen_changed)
		self.window.connect("realize", self.realize_event)
		self.window.connect("enter-notify-event", self.enter_notify_event)
		self.window.connect("leave-notify-event", self.leave_notify_event)
		self.window.connect("focus-in-event", self.focus_in_event)
		self.window.connect("focus-out-event", self.focus_out_event)
		# create menu
		self.menu = gtk.Menu()
		# TEST:
		# if backend shall be connected - do so
		#if init_backend:
		#	self.init_backend()
		#print self.__id
		# /TEST
		# show window
		if show_window:
			self.window.show()
	
	# override for setter-function (Python's version of get/set)
	def __setattr__(self, name, value):
		# set the value in object (ESSENTIAL!!!!)
		object.__setattr__(self, name, value)
		# And do other actions
		if name=="x" or name=="y":
			self.window.move(self.x, self.y)
		elif name == 'scale':
			self.window.resize(int(self.width * self.scale), 
				int(self.height * self.scale))
			# TODO: call on_resize-handler here !!!!
			self.on_scale()
			self.redraw_canvas()
			self.update_shape()
		elif name == "theme_name":
			#self.__dict__ ['theme_name'] = value
			self.load_theme(self.get_theme_dir() + value)
			self.redraw_canvas()
			self.update_shape()
		elif name in ("width", "height"):
			#self.__dict__ [name] = value
			if self.window:
				self.window.resize(self.width, self.height)
				#self.redraw_canvas()
				self.update_shape()
		elif name == "is_widget":
			self.set_is_widget(value)
		elif name == "is_sticky":
			if value == True:
				self.window.stick()
			else:
				self.window.unstick()
		elif name == "keep_above":
			self.window.set_keep_above(bool(value))
			#self.__mi_keep_above.set_active(value)
		elif name == "keep_below":
			self.window.set_keep_below(bool(value))
			#self.__mi_keep_below.set_active(value)
		elif name == "skip_pager":
			if self.window.window:
				self.window.window.set_skip_pager_hint(bool(value))
		elif name == "skip_taskbar":
			if self.window.window:
				self.window.window.set_skip_taskbar_hint(bool(value))
		# TEST: save settings to backend in real-time
		if self.__id:
			o = self.get_option_by_name(name)
			if o != None:
				# save single option to backend
				try:
					self.backend.save_option(self.__id, o.name, 
						o.on_export(value))
				except:
					print "Unable to find backend! Option not saved."
					# disable backend
					self.__id = None
		# /TEST
	
	#-----------------------------------------------------------------------
	# TEST: new settings-backend over dbus-connection
	#-----------------------------------------------------------------------
		
	def close_backend(self):
		"""Close connection with backend."""
		try:
			self.backend.unregister(self.__id)
		except:
			print "Unable to find backend for unregistring!"
		self.__id = None

	def init_backend(self):
		"""Initialize the dbus-connection with the ScreenletsService-daemon."""
		try:
			self.__id = self.backend.register(self.__class__.__name__)
			return True
		except Exception, ex:
			print "Unable to find settings-backend! Options will not be saved."
			self.__id = None
			print ex
			return False
	
	def set_id(self, new_id):
		"""Set new id. This function should never be called!!!!"""
		self.__id = new_id
	
	"""# UNUSED: pass all editable options to the backend (in one call)
	def save_to_backend(self):
		"Save ALL editable options for this Screenlet to the backend."
		lst = []
		for o in self.__options__:
			lst.append([o.name, o.on_export(self.__dict__[o.name])])
		self.backend.save_option_list(self.__id, lst)"""
	
	
	#-----------------------------------------------------------------------
	# Screenlet's public functions
	#-----------------------------------------------------------------------
	
	# NOTE: This function is deprecated and will get removed. The
	# XML-based menus should be preferred
	def add_default_menuitems(self, flags=DefaultMenuItem.STANDARD):
		"""Appends the default menu-items to self.menu. You can add on OR'ed
		   flag with DefaultMenuItems you want to add."""
		# children already exist? add separator
		if len(self.menu.get_children()) > 0:
			self.add_menuitem("", "-")
		# create menu (or submenu?)
		#if flags & DefaultMenuItem.IS_SUBMENU :
		#	menu = gtk.Menu()
		#else:
		menu = self.menu
		# EXPERIMENTAL:
		if flags & DefaultMenuItem.XML:
			# TODO: create XML-menu from screenletpath/menu.xml
			xfile = self.get_screenlet_dir() + "/menu.xml"
			xmlmenu = XmlMenu.create_menu_from_file(xfile, 
				self.menuitem_callback)
			if xmlmenu:
				self.menu = xmlmenu
			pass
		# add size-selection
		if flags & DefaultMenuItem.SIZE:
			size_item = gtk.MenuItem("Size")
			size_item.show()
			size_menu = gtk.Menu()
			menu.append(size_item)
			size_item.set_submenu(size_menu)
			#for i in xrange(10):
			for i in (0.2, 0.5, 0.75, 1.0, 1.5, 2.0, 3.0, 4.0, 5.0, 7.5, 10):
				s = str(int(i * 100))
				item = gtk.MenuItem(s + " %")
				item.connect("activate", self.menuitem_callback, 
					"scale:"+str(i))
				item.show()
				size_menu.append(item)
		# create theme-selection menu
		if flags & DefaultMenuItem.THEMES:
			themes_item = gtk.MenuItem("Theme")
			themes_item.show()
			themes_menu = gtk.Menu()
			menu.append(themes_item)
			themes_item.set_submenu(themes_menu)
			# create theme-list from theme-directory
			lst = self.get_available_themes()
			for tname in lst:
				item = gtk.MenuItem(tname)
				item.connect("activate", self.menuitem_callback, 
					"theme:"+tname)
				item.show()
				themes_menu.append(item)
		# add window-options menu
		if flags & DefaultMenuItem.WINDOW_MENU:
			winmenu_item = gtk.MenuItem("Window")
			winmenu_item.show()
			winmenu_menu = gtk.Menu()
			menu.append(winmenu_item)
			winmenu_item.set_submenu(winmenu_menu)
			# add "Sticky"-menuitem
			self.__mi_sticky = item = gtk.CheckMenuItem("Sticky")
			item.set_active(self.is_sticky)
			item.connect("activate", self.menuitem_callback, 
				"option:sticky")
			item.show()
			winmenu_menu.append(item)
			# add "Widget"-menuitem
			self.__mi_widget = item = gtk.CheckMenuItem("Widget")
			item.set_active(self.is_widget)
			item.connect("activate", self.menuitem_callback, 
				"option:widget")
			item.show()
			winmenu_menu.append(item)
			# add "Keep above"-menuitem
			self.__mi_keep_above = item = gtk.CheckMenuItem("Keep above")
			item.set_active(self.keep_above)
			item.connect("activate", self.menuitem_callback, 
				"option:keep_above")
			item.show()
			winmenu_menu.append(item)
			# add "Keep Below"-menuitem
			self.__mi_keep_below = item = gtk.CheckMenuItem("Keep below")
			item.set_active(self.keep_below)
			item.connect("activate", self.menuitem_callback, 
				"option:keep_below")
			item.show()
			winmenu_menu.append(item)
		# add Settings-item
		if flags & DefaultMenuItem.PROPERTIES:
			self.add_menuitem("", "-")
			self.add_menuitem("options", "Properties...")
		# add info-item
		if flags & DefaultMenuItem.INFO:
			self.add_menuitem("", "-")
			self.add_menuitem("info", "Info...")
		# add delete item
		if flags & DefaultMenuItem.DELETE:
			self.add_menuitem("", "-")
			self.add_menuitem("delete", "Delete Screenlet ...")

	def add_menuitem(self, id, label, callback=None):
		"""Simple way to add menuitems to the right-click menu."""
		if callback == None:
			callback = self.menuitem_callback
		if label == "-":
			menu_item = gtk.SeparatorMenuItem()
		else:
			menu_item = gtk.MenuItem(label)
			menu_item.connect("activate", callback, id)
		self.menu.append(menu_item)
		menu_item.show()
		return menu_item
	
	def clear_cairo_context(self, ctx):
		"""Fills the given cairo.Context with fully transparent white."""
		ctx.save()
		ctx.set_source_rgba(1, 1, 1, 0)
		ctx.set_operator (cairo.OPERATOR_SOURCE)
		ctx.paint()
		ctx.restore()

	def close(self):
		"""Close this Screenlet
		   TODO: send close-notify instead of destroying window?"""
		#self.save_settings()
		self.window.unmap()
		self.window.destroy()
		
	def get_screenlet_dir(self):
		"""Return the name of this screenlet's personal directory."""
		if self.__path__ != '':
			return self.__path__
		else:
			return os.getcwd()
	
	def get_theme_dir(self):
		"""Return the name of this screenlet's personal theme-dir."""
		return self.get_screenlet_dir() + "/themes/"
	
	def get_available_themes(self):
		"""Returns a list with the names of all available themes in this
			Screenlet's theme-directory."""
		dirname = self.get_theme_dir()
		dirlst = glob.glob(dirname + '*')
		dirlst.sort()
		tdlen = len(dirname)
		lst = []
		for fname in dirlst:
			dname = fname[tdlen:]
			# TODO: check if it's a dir
			lst.append(dname)
		return lst
	
	def hide(self):
		"""Hide this Screenlet's underlying gtk.Window"""
		self.window.hide()
		self.on_hide()
	
	# EXPERIMENTAL:
	# NOTE: load_theme does NOT call redraw_canvas and update_shape!!!!!
	# To do all in one, set attribute self.theme_name instead
	#def on_load_theme(self): pass
	def load_theme(self, path):
		"""Load a theme for this Screenlet from the given path. NOTE: 
load_theme does NOT call redraw_canvas and update_shape!!!!! To do all in one, 
set the attribute self.theme_name instead."""
		if self.theme:
			self.theme.free()
			del self.theme
		self.theme = ScreenletTheme(path)
		# check for errors
		if self.theme.loaded == False:
			print "Error while loading theme: " + path
			self.theme = None
		else:
			# call user-defined handler
			self.on_load_theme()
			# if override options is allowed, apply them
			if self.allow_option_override:
				if self.theme.has_overrides():
					if show_question(self, 
						'This theme wants to override your settings for '+\
						'this Screenlet. Do you want to allow that?') == False:
						return
					self.theme.apply_option_overrides(self)
	# /EXPERIMENTAL
	
	def main(self):
		"""If the Screenlet runs as stand-alone app, starts gtk.main()"""
		gtk.main()
	
	def set_is_widget(self, value):
		"""Set this window as a Widget (or not)"""
		if value:
			# set window type to utility
			self.window.window.set_type_hint(
				gtk.gdk.WINDOW_TYPE_HINT_UTILITY)
			# set _compiz_widget-property
			self.window.window.property_change("_COMPIZ_WIDGET", 
				gtk.gdk.SELECTION_TYPE_WINDOW,
				32, gtk.gdk.PROP_MODE_REPLACE, (True,))
		else:
			# set window type to normal
			self.window.window.set_type_hint(
				gtk.gdk.WINDOW_TYPE_HINT_NORMAL)
			# set _compiz_widget-property
			self.window.window.property_delete("_COMPIZ_WIDGET")
		# notify handler
		self.on_switch_widget_state(value)
	
	def show(self):
		"""Show this Screenlet's underlying gtk.Window"""
		self.window.show()
		self.window.move(self.x, self.y)
		self.on_show()
	
	def show_about_dialog(self):
		"""Show a simple About-Dialog for this Screenlet (by Sorcerer)"""
		#meta = self.__metainfo
		aboutdialog = gtk.AboutDialog()
		aboutdialog.set_name(self.__name__)
		aboutdialog.set_version(self.__version__)
		aboutdialog.set_comments(self.__desc__)
		aboutdialog.set_copyright("(c) 2007 by "+self.__author__)
		#aboutdialog.show()
		aboutdialog.run()
		aboutdialog.destroy()
		
	def show_settings_dialog(self):
		"""Show the EditableSettingsDialog for this Screenlet."""
		se = OptionsDialog(490, 450)
		img = gtk.Image()
		try:
			icn = gtk.gdk.pixbuf_new_from_file(self.get_screenlet_dir() + 
				'/icon.svg')
			img.set_from_pixbuf(icn)
		except:
			img.set_from_stock(gtk.STOCK_PROPERTIES, 5)
		se.set_title('Options: ' + self.__name__)
		se.set_info(self.__name__, self.__desc__, '(c) by '+self.__author__, 
			version='v'+self.__version__, icon=img)
		se.show_options_for_object(self)
		resp = se.run()
		if resp == gtk.RESPONSE_REJECT:	# TODO!!!!!
			se.reset_to_defaults()
		else:
			self.update_shape()
			# TEST
			# export options and send them to backend (as list with strings)
			#self.save_to_backend()
			# /TEST
		se.destroy()
	
	def redraw_canvas(self):
		"""Redraw the Screenlet."""
		if self.window:
			alloc = self.window.get_allocation()
			rect = gtk.gdk.Rectangle(alloc.x, alloc.y, 
				alloc.width, alloc.height)
			if self.window.window:
				self.window.window.invalidate_rect(rect, True)
				self.window.window.process_updates(True)
	
	def update_shape(self):
		"""Update window shape (only call this when shape has changed).
			TODO: cache the bitmap and don't create new if size is unchanged"""
		w = int(self.width * self.scale)
		h = int(self.height * self.scale)
		if w==0:
			w = 100
		if h==0:
			h = 100
		"""data = ""
		for i in xrange(w * h):
			data += "0" """
		data = ''.zfill(w*h)
		bitmap = gtk.gdk.bitmap_create_from_data(None, data, w, h)
		ctx = bitmap.cairo_create()
		self.clear_cairo_context(ctx)		#TEST
		self.on_draw_shape(ctx)
		self.window.input_shape_combine_mask(bitmap, 0, 0)
	
	# ----------------------------------------------------------------------
	# Screenlet's event-handler dummies
	# ----------------------------------------------------------------------
	
	def on_delete(self, event):
		"""Called when the Screenlet gets deleted. Return True to cancel."""
		return False
	
	def on_draw(self, ctx):
		"""Callback for drawing the Screenlet's window - override
in subclasses to implement your own drawing."""
		pass
	
	def on_draw_shape(self, ctx):
		"""Callback for drawing the Screenlet's shape - override
in subclasses to draw the window's input-shape-mask."""
		pass
		
	def on_focus(self, event):
		"""Called when the Screenlet's window receives focus."""
		pass
	
	def on_hide(self):
		"""Called when the Screenlet gets hidden."""
		pass

	def on_load_theme(self):
		"""Called when the theme is reloaded (after loading, before redraw)."""
		pass
	
	def on_menuitem_select(self, id):
		"""Called when a menuitem is selected."""
		pass
	
	def on_mouse_down(self, event):
		"""Called when a buttonpress-event occured in Screenlet's window. 
Returning True causes the event to be not further propagated."""
		return False
	
	def on_mouse_enter(self, event):
		"""Called when the mouse enters the Screenlet's window."""
		pass
		
	def on_mouse_leave(self, event):
		"""Called when the mouse leaves the Screenlet's window."""
		pass
	
	def on_mouse_up(self, event):
		"""Called when a buttonrelease-event occured in Screenlet's window. 
Returning True causes the event to be not further propagated."""
		return False
	
	def on_quit(self):
		"""Callback for handling destroy-event"""
		return True
		
	def on_realize(self):
		""""Callback for handling realize-event."""
	
	def on_scale(self):
		"""Called when Screenlet.scale is changed."""
		pass
	
	def on_show(self):
		"""Called when the Screenlet gets shown after being hidden."""
		pass
	
	def on_switch_widget_state(self, state):
		"""Called when the Screenlet enters/leaves "Widget"-state."""
		pass
	
	def on_unfocus(self, event):
		"""Called when the Screenlet's window loses focus."""
		pass
	
	# ----------------------------------------------------------------------
	# Screenlet's event-handlers for GTK-events
	# ----------------------------------------------------------------------
	
	def alpha_screen_changed(self, window, screen=None):
		"""set colormap for window"""
		if screen==None:
			screen = window.get_screen()
		map = screen.get_rgba_colormap()
		if map:
			pass
		else:
			map = screen.get_rgb_colormap()
		window.set_colormap(map)		
	
	def button_press(self, widget, event):
		# sett flags for user-handler
		if event.button == 1:
			self.is_dragged = True
		# call user-handler for onmousedown
		if self.on_mouse_down(event) == True:
			return True
		# unhandled? continue
		if event.button == 1:
			widget.begin_move_drag(event.button, 
				int(event.x_root), int(event.y_root), event.time)
		elif event.button == 3:
			self.menu.popup(None, None, None, event.button, 
				event.time)
		return False
	
	def button_release(self, widget, event):
		#print "Button release"
		self.is_dragged = False
		if self.on_mouse_up(event):
			return True
		return False
	
	# NOTE: this is a nasty workaround because I was unable
	# to find the right event for getting notified about
	# end of a gtk.Window.begin_move_drag(). And I think the problem
	# is the move_drag-operation itself ...
	def configure_event(self, widget, event):
		#print "onConfigure"
		self.is_dragged = False
		# update position
		pos = self.window.get_position()
		if pos[0] != self.x:
			self.x = pos[0]
		if pos[1] != self.y:
			self.y = pos[1]
		return False
	
	def delete_event(self, widget, event, data=None):
		# cancel event?
		if self.on_delete(event) == True:
			print "Cancel event"
			return True
		return False

	def destroy(self, widget, data=None):
		# call user-defined on_quit-handler
		self.on_quit()
		#print "destroy signal occurred"
		self.emit("screenlet_removed", self)
		# close backend
		if self.__id:
			self.close_backend()	
		# close gtk?
		if self.quit_on_close:
			gtk.main_quit()
		else:
			del self
	
	def drag_end(self, widget, drag_context):
		print "End drag"
		self.is_dragged = False
		return False
	
	def drag_begin(self, widget, drag_context):
		print "Start drag"
		return False
	
	def enter_notify_event(self, widget, event):
		#self.__mouse_inside = True
		self.on_mouse_enter(event)
		#self.redraw_canvas()
	
	def expose(self, widget, event):
		ctx = widget.window.cairo_create()
		# clear context
		self.clear_cairo_context(ctx)
		# set a clip region for the expose event
		ctx.rectangle(event.area.x, event.area.y,
			event.area.width, event.area.height)
		ctx.clip()
		# scale context
		#ctx.scale(self.scale, self.scale)
		# call drawing method
		self.on_draw(ctx)
		# and delete context
		del ctx
		return False
	
	def focus_in_event(self, widget, event):
		self.on_focus(event)
	
	def focus_out_event(self, widget, event):
		self.on_unfocus(event)
		
	def leave_notify_event(self, widget, event):
		#self.__mouse_inside = False
		self.on_mouse_leave(event)
	
	def menuitem_callback(self, widget, id):
		if id == "delete":
			self.close()
		elif id in ("info", "about"):
			self.show_about_dialog()
		elif id in ("settings", "options", "properties"):
			# show settings dialog
			self.show_settings_dialog()
		elif id.startswith('scale:'):
			self.scale = float(id[6:])
			self.update_shape()
		elif id[:5] == "size:":
			# set size and update shape (redraw is done by setting height)
			#self.__dict__['width'] = int(id[5:])
			self.width = int(id[5:])
			self.height = int(id[5:])
			self.update_shape()
		elif id[:6]=="theme:":
			print "Screenlet: Set theme " + id[6:]
			# set theme
			self.theme_name = id[6:]
		elif id[:8] == "setting:":
			# set a boolean option to the opposite state
			try:
				if type(self.__dict__[id[8:]]) == bool:
					self.__dict__[id[8:]] = not self.__dict__[id[8:]]	# UNSAFE!!
			except:
				print "Error: Cannot set missing or non-boolean value '"\
					+ id[8:] + "'"
		elif id[:7] == "option:":
			# NOTE: this part should be removed and XML-menus
			#		should be used by default ... maybe
			# set option
			if id[7:]=="sticky":
				self.is_sticky = not self.is_sticky
				#widget.toggle()
			elif id[7:]=="widget":
				self.is_widget = not self.is_widget
			elif id[7:]=="keep_above":
				self.keep_above = not self.keep_above
				self.__mi_keep_above.set_active(self.keep_above)
			elif id[7:]=="keep_below":
				self.keep_below = not self.keep_below
				self.__mi_keep_below.set_active(self.keep_below)
		else:
			#print "Item: " + string
			pass
		# call simplified handler
		self.on_menuitem_select(id)
		return False
	
	def realize_event(self, widget):
		"""called when window has been realized"""
		if self.window.window:
			self.window.window.set_back_pixmap(None, False)
		if self.is_widget:
			self.set_is_widget(True)
		self.on_realize()
	
	


# TEST!!!
class ShapedWidget(gtk.DrawingArea):
	"""A simple base-class for creating owner-drawn gtk-widgets"""
	
	__widget=None
	
	mouse_inside = False
	width = 32
	height = 32
	
	def __init__(self, width, height):
		# call superclass
		super(ShapedWidget, self).__init__()
		# create/setup widget
		#self.__widget = gtk.Widget()
		self.set_app_paintable(True)
		self.set_size_request(width, height)
		# connect handlers
		self.set_events(gtk.gdk.ALL_EVENTS_MASK)
		self.connect("expose-event", self.expose_event)
		self.connect("button-press-event", self.button_press)
		self.connect("button-release-event", self.button_release)
		self.connect("enter-notify-event", self.enter_notify)
		self.connect("leave-notify-event", self.leave_notify)
	
	# EXPERIMENTAL: TODO: cache bitmap until size changes
	def update_shape(self):
		"""update widget's shape (only call this when shape has changed)"""
		data = ""
		for i in xrange(self.width*self.height):
			data += "0"
		bitmap = gtk.gdk.bitmap_create_from_data(None, 
			data, self.width, self.height)
		ctx = bitmap.cairo_create()
		ctx.set_source_rgba(1, 1, 1, 0)
		ctx.set_operator (cairo.OPERATOR_SOURCE)
		ctx.paint()
		self.draw_shape(ctx)
		self.input_shape_combine_mask(bitmap, 0, 0)
		print "Updating shape."
	
	def button_press(self, widget, event):
		if event.button==1:
			print "left button pressed!"
		return False
		
	def button_release(self, widget, event):
		#if event.button==1:
			#print "left button release!"
		return False
	
	def enter_notify(self, widget, event):
		self.mouse_inside = True
		self.queue_draw()
		#print "mouse enter"
	
	def leave_notify(self, widget, event):
		self.mouse_inside = False
		self.queue_draw()
		#print "mouse leave"
	
	def draw(self, ctx):
		pass
		
	def draw_shape(self, ctx):
		self.draw(ctx)
	
	def expose_event(self, widget, event):
		ctx = widget.window.cairo_create()
		# set a clip region for the expose event
		ctx.rectangle(event.area.x, event.area.y,
			event.area.width, event.area.height)
		ctx.clip()
		# clear context
		ctx.set_source_rgba(1, 1, 1, 0)
		ctx.set_operator (cairo.OPERATOR_SOURCE)
		ctx.paint()
		# call drawing method
		self.draw(ctx)
		# and delete context
		del ctx
		return False


# TEST (as the name implies)
"""class TestWidget(ShapedWidget):
	
	def __init__(self, width, height):
		#ShapedWidget.__init__(self, width, height)
		super(TestWidget, self).__init__(width, height)
	
	def draw(self, ctx):
		if self.mouse_inside:
			ctx.set_source_rgba(1, 0, 0, 0.8)
		else:
			ctx.set_source_rgba(1, 1, 0, 0.8)
		ctx.rectangle(0, 0, 32, 32)
		ctx.fill()
"""


# ------------------------------------------------------------------------------
# MODULE-FUNCTIONS
# ------------------------------------------------------------------------------

def create_new_instance(classname):
	""""Create a new Screenlet-instance for the given Screenlet-classname.
TODO: create new through dbus-call if backend is running"""
	exec "from "+classname+" import "+classname
	classobj = eval(classname)
	if classobj != None:
		sl = classobj()
		if sl:
			sl.init_backend()
			return sl
	return None

# a very basic module loader function (with support for passing args)
def load_class_from_module(modpath, modname, classname, **keyword_args):
	"""Import module from path and return an instance of classname. Does 
	NOT allow underscores in classnames!"""
	# security check (to avoid passing unwanted stuff to exec)
	if False in (classname.isalnum(), modname.isalnum()):
		print "load_class_from_module: Illegal modname or classname."
		return None
	if sys.path.count(modpath) == 0:
		sys.path.insert(0, modpath)
	try:
		exec "import " + modname
		#exec "mod=" + modname + "." + classname + "(path='"+path+"')"
		exec "mod=" + modname + "." + classname + "(**keyword_args)"
		return mod
	except Exception, ex:
		print "Unable to load '" + modpath +"/" + modname + "." + classname + \
			"' - " + str(ex)
		return None

def import_and_create_screenlet(classname, **keyword_args):
	"""Look through all paths in SCREENLETS_PATH and search for the
first occurence of the given screenlets' directory. When found, 
the Screenlet gets imported and instanced, else None is returned."""
	lpath = classname[:-9]#.lower()
	print "import_and_create_screenlet: "+classname
	for path in SCREENLETS_PATH:
		#print path
		plen = len(path) + 1
		files = glob.glob(path + '/*')
		for fi in files:
			cur = fi[plen:]
			if cur == lpath:
				#print "FOUND!!"
				# found path, import and instantiate the new Screenlet
				sl = load_class_from_module(path + '/' + lpath, 
					classname, classname, path=path + '/' + lpath, 
					**keyword_args)
				if sl != None:
					return sl
				else:
					return False
					
def show_message(screenlet, message):
	"""Show a message for the given Screenlet (may contain Pango-Markup)."""
	md = gtk.MessageDialog(screenlet.window, type=gtk.MESSAGE_INFO, 
		buttons=gtk.BUTTONS_OK)
	md.set_title(screenlet.__name__)
	md.set_markup(message)
	md.run()
	md.destroy()

def show_question(screenlet, message):
	"""Show a question for the given Screenlet (may contain Pango-Markup)."""
	md = gtk.MessageDialog(screenlet.window, type=gtk.MESSAGE_QUESTION, 
		buttons=gtk.BUTTONS_YES_NO)
	md.set_title(screenlet.__name__)
	md.set_markup(message)
	response = md.run()
	md.destroy()
	if response == gtk.RESPONSE_YES:
		return True
	return False

def show_error(screenlet, message):
	"""Show an error for the given Screenlet (may contain Pango-Markup)."""
	md = gtk.MessageDialog(screenlet.window, type=gtk.MESSAGE_ERROR,
		buttons=gtk.BUTTONS_OK)
	md.set_title('Error in ' + screenlet.__name__)
	md.set_markup(message)
	md.run()
	md.destroy()
