#!/usr/bin/env python

#  ControlScreenlet (c) RYX 2007 <ryx@ryxperience.com>
#
# INFO:
# - a simple "control" for adding Screenlets and a possible replacement
#   for the system-menu
#
# TODO:
# - ...

import screenlets
from screenlets import DefaultMenuItem
from screenlets.options import BoolOption
import cairo
import os


class ControlScreenlet(screenlets.Screenlet):
	"""A simple "control" for adding/managing Screenlets. It
	   cares for saving/restoring settings/positions of Screenlets."""
	
	# default meta-info for Screenlets
	__name__ = 'ControlScreenlet'
	__version__ = '0.4'
	__author__ = 'RYX (aka Rico Pfaus)'
	__desc__ = 'A simple "control" for adding Screenlets to the backend.'
		
	# menuitems
	menuitem_add = None
	#menuitem_hideshow = None
	
	_hidden = False
	
	# editable settings
	add_screenlet_as_widget = False
	#hide_show_on_click = False
	
	# constructor
	def __init__(self, **keyword_args):
		# call super
		screenlets.Screenlet.__init__(self, is_widget=False, 
			is_sticky=True, uses_theme=True, **keyword_args)
		# set theme and redraw
		self.theme_name = "default"
		# create menu
		self.add_default_menuitems(DefaultMenuItem.XML)
		# quit gtk when this window is closed
		self.quit_on_close = True
		# add editable settings
		self.add_options_group('Control', 
			'Additional settings for the ControlScreenlet.')
		# add editable settings to this Screenlet
		self.add_option(BoolOption('Control', 'add_screenlet_as_widget', 
			self.add_screenlet_as_widget, 'Add as "Widget"', 
			'If active, adds new Screenlets as "Widgets" (NOTE: When this ' + 
			'is active, you will not immediately see newly added Screenlets ' +
			'if you are not in "Widget"-mode (which depends on your ' +
			'Windowmanager and is only fully supported by compiz yet))'))
		"""self.add_option(BoolOption('Control', 'hide_show_on_click', 
			self.hide_show_on_click, 'Hide/Show on click', 
			'Hide/Show all open Screenlets (added by this Control) when '+
			'the Control is left-clicked.'))"""

	def on_mouse_down(self, event):
		"""If hide_show_on_click is active, switch hide/show state."""
		#if self.hide_show_on_click and event.button==1:
		#	#self.switch_hide_show()
		#	return True	
		pass
	
	def on_menuitem_select(self, id):
		"""handle MenuItem-events in right-click menu"""
		#TEST
		#if id=="test":
		#print "ITEM: "+id
		#/TEST
		if id == "hide_show":
			# TODO: use DBus-call for this
			#self.switch_hide_show()
			pass
		elif id[:4] == "add:":
			# make first letter uppercase (workaround for xml-menu)
			cls = id[4].upper() + id[5:]
			#cls = id[4:]
			#print "Add: "+cls
			# add screenlet to backend (by its classname)
			try:
				self.backend.add(cls, is_widget=self.add_screenlet_as_widget)
			except:
				screenlets.show_error(self, 'No backend running - ' + \
					'failed to add Screenlet.')
		elif id[:5] == "exec:":
			# execute shell command
			os.system(id[5:] + " &")
		elif id == "delete":
			# TODO: ask before quit
			#gtk.main_quit()
			self.close()
			del self
			return
	
	def on_draw(self, ctx):
		# set scale
		ctx.scale(self.scale, self.scale)
		ctx.set_operator(cairo.OPERATOR_OVER)
		# render bg
		if self.theme:
			if self._hidden:
				self.theme['control-bg-hidden.svg'].render_cairo(ctx)
			else:
				self.theme['control-bg.svg'].render_cairo(ctx)
	
	def on_draw_shape(self, ctx):
		#ctx.scale(self.width/200.0, self.height/200.0)
		ctx.scale(self.scale, self.scale)
		if self.theme:
			self.theme['control-bg.svg'].render_cairo(ctx)
		pass

	
# If the program is run directly or passed as an argument to the python
# interpreter then create a Screenlet instance and show it
if __name__ == "__main__":
	sl = screenlets.create_new_instance('ControlScreenlet')
	sl.main()
		
