#!/usr/bin/env python

#  ExampleScreenlet (c) RYX 2007 <ryx@ryxperience.com>
#
# INFO:
# - a simple example for creating a Screenlet
# 
# TODO:
# - make a nice Screenlet from this example

import screenlets
from screenlets.options import StringOption
import cairo

class ExampleScreenlet(screenlets.Screenlet):
	"""A simple example of how to create a Screenlet"""
	
	# default meta-info for Screenlets
	__name__ = 'ExampleScreenlet'
	__version__ = '0.2'
	__author__ = 'RYX'
	__desc__ = 'A simple example of how to create a Screenlet. The example '+\
		'graphic is NOT my logo and obviously an ugly approach of creating '+\
		'SVG-based graffiti-styles :) ... (my skills are a little "rusted").'

	# editable options
	test_text = ''
		
	# constructor
	def __init__(self, **keyword_args):
		#call super
		screenlets.Screenlet.__init__(self, width=500, height=200, 
			**keyword_args)
		# set theme
		self.theme_name = "default"
		# add default menu items
		self.add_default_menuitems()
		# add option group
		self.add_options_group('Example', 'This is an example of editable' +\
			' options within an options-group ...')
		# add editable option
		self.add_option(StringOption('Example', # group name
			'test_text', 						# attribute-name
			self.test_text,						# default-value
			'Test-Text', 						# widget-label
			'The Test-Text option for this Screenlet ...'	# description
			)) 
	
	def on_draw(self, ctx):
		# if theme is loaded
		if self.theme:
			# set scale rel. to scale-attribute
			ctx.scale(self.scale, self.scale)
			# render svg-file
			self.theme['example-bg.svg'].render_cairo(ctx)
			# render png-file
			ctx.set_source_surface(self.theme['example-test.png'], 0, 0)
			ctx.paint()
	
	def on_draw_shape(self, ctx):
		self.on_draw(ctx)
	
# If the program is run directly or passed as an argument to the python
# interpreter then create a Screenlet instance and show it
if __name__ == "__main__":
	sl = screenlets.create_new_instance('ExampleScreenlet')
	sl.main()

