#!/usr/bin/env python

#  LauncherScreenlet (c) RYX (aka Rico Pfaus) 2007 <ryx@ryxperience.com>
#
# INFO:
# - a Screenlet that launches an application when double-clicked
# 
# TODO:
# - scale window to icon-size (100/icon_size)
# - width/height not working when set through option-editor (strange)
# - optional Label-attribute
#

import screenlets
from screenlets import DefaultMenuItem
from screenlets.options import StringOption, FileOption
import cairo
import rsvg
import gtk
import os


class LauncherScreenlet(screenlets.Screenlet):
	"""A Screenlet that executes any kind of shell-command when clicked."""
	
	# default meta-info for Screenlets
	__name__ = 'LauncherScreenlet'
	__version__ = '0.3'
	__author__ = 'RYX (aka Rico Pfaus)'
	__desc__ = 'A Screenlet that executes a shell-command when clicked.'+\
			   'You can execute any type of shell-command through '+\
			   'this Screenlet (e.g. it offers the ability to "dbus-send").'
	
	# internals
	__svgh_icon = None
	__mouse_inside = False
	__left_mouse_down = False
	__icon_width = 32
	__icon_height = 32
	
	# editable options
	#action = "dbus-send --type=method_call --dest=org.freedesktop.compiz " + \
	#		"/org/freedesktop/compiz/rotate/allscreens/rotate_right " + \
	#		"org.freedesktop.compiz.activate string:'root' int32:0"
	action = ''
	icon = ''
	
	# constructor
	def __init__(self, **keyword_args):
		#call super
		screenlets.Screenlet.__init__(self, **keyword_args)
		# add default menu items
		self.add_default_menuitems(DefaultMenuItem.SIZE |
			DefaultMenuItem.WINDOW_MENU | DefaultMenuItem.PROPERTIES |
			DefaultMenuItem.DELETE)
		# set default icon and action
		self.icon = self.__path__ + '/default-icon.svg'
		#self.action = LauncherScreenlet.action
		# add editable settings
		self.add_options_group('Starter', 
			'Some options related to the Launcher-Screenlet.')
		self.add_option(StringOption('Starter', 'action', 
			self.action, 'Command', 
			'Shell-command to be executed when Icon is clicked ...'))
		self.add_option(FileOption('Starter', 'icon', 
			os.environ['PWD']+'/launcher/default-icon.svg', 'Icon-Filename', 
			'Absolute path to SVG-icon for this Launcher ...'))
		
	def __setattr__(self, name, value):
		screenlets.Screenlet.__setattr__(self, name, value)
		if name=="icon" and value != None and value != '':
			try:
				self.__svgh_icon = rsvg.Handle(value)
				if self.__svgh_icon:
					icon_size = self.__svgh_icon.get_dimension_data()
					self.__icon_width = icon_size[0]
					self.__icon_height = icon_size[0]
					"""if self.window:
						# calculate new bas size relative to icon size
						self.__dict__['height'] = \
							int((self.height * icon_size[0]) / 100.0)
						self.width = int((self.width * icon_size[1]) / 100.0)
						self.update_shape()"""
					if self.window:
						self.update_shape()
						self.redraw_canvas()
			except:
				screenlets.show_message(self, 'Error - ' + 
					'failed to load icon "' + str(self.icon) + '"')
				import sys
				sys.exit(15)
		elif name=="action":
			tt = gtk.Tooltips()
			tt.set_tip(self.window, 'Action: "' + value + '"')
	
	def launch_action(self):
		"""Execute the action/command for this LauncherScreenlet."""
		if self.action:
			os.system(self.action + "&")
		else:
			screenlets.show_message(self, "You have first to define a " +
				"command to be launched for this LauncherScreenlet (see " + 
				"Options/Starter/Command in this Screenlet's right-click " +
				"menu).")
	
	def on_mouse_enter(self, event):
		"""Mouse entered the Screenlet."""
		self.__mouse_inside = True
		self.redraw_canvas()
	
	def on_mouse_leave(self, event):
		"""Mouse left the Screenlet."""
		self.__mouse_inside = False
		self.__left_mouse_down = False
		self.redraw_canvas()
	
	def on_mouse_down(self, event):
		"""Mouse pressed on the Screenlet."""
		if event.button == 1:
			self.__left_mouse_down = True
			self.redraw_canvas()
			self.launch_action()
			return True
	
	def on_mouse_up(self, event):
		"""Mouse released on the Screenlet."""
		if event.button == 1:
			self.__left_mouse_down = False
			self.redraw_canvas()
			
	def on_draw(self, ctx):
		"""Draw the Screenlet's graphics."""
		#ctx.scale(self.width/self.__icon_width, self.height/self.__icon_height)
		ctx.scale(self.scale, self.scale)
		ctx.set_operator(cairo.OPERATOR_OVER)
		if self.__svgh_icon:
			self.__svgh_icon.render_cairo(ctx)
			if self.__mouse_inside:
				if self.__left_mouse_down:
					ctx.set_source_rgba(0, 0, 0, 0.3)
				else:
					ctx.set_source_rgba(1, 1, 1, 0.3)
				ctx.set_operator(cairo.OPERATOR_ATOP)
				ctx.rectangle(0, 0, self.width, self.height)
				ctx.fill()
		
	def on_draw_shape(self, ctx):
		"""Draw the Screenlet's input shape."""
		if self.__svgh_icon:
			self.on_draw(ctx)

	
# If the program is run directly or passed as an argument to the python
# interpreter then create a Screenlet instance and show it
if __name__ == "__main__":
	sl = screenlets.create_new_instance('LauncherScreenlet')
	sl.main()
	
