#!/usr/bin/env python

#  MailCheckScreenlet (c) RYX 2007 <ryx@ryxperience.com>
#
# INFO:
# - A Screenlet to check for new mails on a POP3-server.
# 
# TODO:
# - add unchecked-count instead of bool-flag
# - SSL-support
# - password-encrytion
# - send email by dragging text onto the Screenlet

import screenlets
from screenlets.options import StringOption, IntOption
import gobject
import cairo
import pango
import socket
import poplib
import os
from gtk import Tooltips

class MailCheckStatus:
	REFRESH = 1
	GOT_MAIL = 2
	ERROR = 3
	IDLE = 100


class MailCheckScreenlet(screenlets.Screenlet):
	"""A Screenlet to check for new mails on a POP3-server."""
	
	# default meta-info for Screenlets
	__name__ = 'MailCheckScreenlet'
	__version__ = '0.1'
	__author__ = 'RYX (aka Rico Pfaus) 2007'
	__desc__ = 'A Screenlet that notifies you about new e-mail on a '+\
		'POP3-server. When you receive new mail, the icon starts flashing .'+\
		'and you get quick access to your favorite eMail-client program.'+\
		'<span color="red">WARNING: Password is saved as plain text!!!!</span>'
	
	# internals
	#__mail_count = 0
	__timeout = None
	__status = MailCheckStatus.IDLE
	__unchecked = False
	__flash_phase = 0
	__flash_timeout = None
	__tooltips = Tooltips()
	
	# editable options
	check_interval = 1		# minutes!!!
	server_url = ''
	server_user = ''
	server_pass = ''
	mail_client = 'evolution'
	mail_count = 0		# hidden option to remember number of mails

	# constructor
	def __init__(self, **keyword_args):
		# call super
		screenlets.Screenlet.__init__(self, **keyword_args)
		# set theme
		self.theme_name = "default"
		# add menuitems
		self.add_menuitem('check_mail', 'Check now!')
		self.add_menuitem('open_client', 'Open client ...')
		self.add_default_menuitems()
		# add option group
		self.add_options_group('E-Mail', 'POP3/Mail-related options ...')
		# add editable options
		self.add_option(IntOption('E-Mail','check_interval', 
			self.check_interval, 'Checking interval (minutes)', 
			'The interval (in minutes) after that is checked for new mail ...', 
			min=1, max=1200))
		self.add_option(StringOption('E-Mail','server_url', self.server_url,
			'POP3-Server URL', 'The url of the POP3-server to check ...'), 
			realtime=False)
		self.add_option(StringOption('E-Mail','server_user', self.server_user,
			'POP3-Username', 'The username for the POP3-mailbox to check ...'))
		self.add_option(StringOption('E-Mail','server_pass', self.server_pass,
			'POP3-Password', 'The password for the POP3-mailbox to check ...'))
		self.add_option(StringOption('E-Mail','mail_client', self.mail_client,
			'Mail-Client', 'The e-mail client-application to open ...'))
		self.add_option(IntOption('E-Mail','mail_count', self.mail_count,
			'', '', hidden=True))
		# initial check:
		gobject.timeout_add(500, self.initial_check)
		
	def __setattr__(self, name, value):
		screenlets.Screenlet.__setattr__(self, name, value)
		if name == 'check_interval':
			self.set_check_interval(value * 60000)
		elif name == 'server_url':
			gobject.idle_add(self.initial_check)
		
	def initial_check(self):
		"""Initially checks for mail and then starts the timeout."""
		self.check_email()
		self.set_check_interval(self.check_interval * 60000)
		return False
	
	def check_email(self):
		"""Check the current server for email (should be run as thread)."""
		# set default timeout for all socket connections to 30 secs
		socket.setdefaulttimeout(30000)
		print "Connecting to POP3-server ... please wait."
		self.__status = MailCheckStatus.REFRESH
		self.redraw_canvas()
		try:
			server = poplib.POP3(self.server_url)
		except:
			screenlets.show_error(self, "Error while connecting to server.")
			self.__status = MailCheckStatus.ERROR
			self.redraw_canvas()
			return False
		# authenticate
		try:
			print server.user(self.server_user)
			print server.pass_(self.server_pass)
		except:
			screenlets.show_error(self, "Invalid login data given.")
			self.__status = MailCheckStatus.ERROR
			self.redraw_canvas()
			return False
		# get list with mails (response, list-of-mails)
		resp = server.list()
		if resp[0].startswith('+OK'):
			messages = resp[1]
			msgnum = len(messages)
			if msgnum > self.mail_count:
				diff = msgnum-self.mail_count
				self.mail_count = msgnum
				self.__status = MailCheckStatus.GOT_MAIL
				#print "GOT_MAIL"
				self.__unchecked = True
				self.__flash_timeout = gobject.timeout_add(500, 
					self.flash_interval)
				tipstr = "You have new mails ("+str(diff)+"/"+str(diff)+")"
			elif msgnum <= self.mail_count:
				self.mail_count = msgnum
				self.__status = MailCheckStatus.IDLE
				tipstr = "You have no new mails (0/"+str(msgnum)+")"
				#print "IDLE"
			self.__tooltips.set_tip(self.window, tipstr)
		else:
			screenlets.show_error(self, "Unable to retrieve mails from server.")
			self.__status = MailCheckStatus.ERROR
			self.redraw_canvas()
			return False
		# close connection
		server.quit()
		# redraw canvas
		self.redraw_canvas()
		# return True (to continue timeout)
		return True
	
	def set_check_interval(self, interval):
		"""set the interval time for checking mails (in milliseconds)."""
		#print "check_interval"
		if self.__timeout:
			gobject.source_remove(self.__timeout)
		self.__timeout = gobject.timeout_add(interval, self.check_email)	
	
	def flash_interval(self):
		"""Timeout-interval function for flashing-effect"""
		#print "flash_interval"
		if self.__status == MailCheckStatus.GOT_MAIL:
			self.redraw_canvas()
			return True
		return False
	
	def set_idle(self):
		"""Reset the status back to idle state. (BUGGY)"""
		if self.__flash_timeout:
			gobject.source_remove(self.__flash_timeout)
		self.__flash_phase = 0.0
		self.__unchecked = False
		self.__status == 100#MailCheckStatus.IDLE
		self.redraw_canvas()
			
	def on_menuitem_select(self, id):
		if id == 'check_mail':
			#self.__unchecked = False
			gobject.idle_add(self.initial_check)
		elif id == 'open_client':
			# TODO: execute e-mail application
			#print "Open client"
			self.set_idle()
			if self.mail_client != '':
				os.system(self.mail_client + ' &')
			else:
				screenlets.show_error(self, "You have to define an "+\
					"eMail-client to be opened by this action (e.g. "+\
					"evolution or thunderbird). Go to the Properties in "+\
					"the right-click menu to define your client of choice.")
	
	def on_draw(self, ctx):
		ctx.scale(self.scale, self.scale)
		if self.theme:
			#ctx.scale(self.width/self.theme.width,
			#	self.height/self.theme.height)
			#print "DRAW: "+str(self.__status)
			if self.__status == MailCheckStatus.GOT_MAIL or \
				self.__unchecked == True:
				self.theme['mailcheck-got-mail.svg'].render_cairo(ctx)
			elif self.__status == MailCheckStatus.IDLE:
				self.theme['mailcheck-bg.svg'].render_cairo(ctx)
			elif self.__status == MailCheckStatus.REFRESH:
				self.theme['mailcheck-refresh.svg'].render_cairo(ctx)
			elif self.__status == MailCheckStatus.ERROR:
				self.theme['mailcheck-error.svg'].render_cairo(ctx)
		else:
			# TEMP: fallback bg
			ctx.rectangle(0, 0, self.width, self.height)
			ctx.set_source_rgba(1, 0, 0, 0.8)
			ctx.fill()
		# flashing?
		if self.__status == MailCheckStatus.GOT_MAIL:
			ctx.save()
			ctx.set_operator(cairo.OPERATOR_ATOP)
			ctx.rectangle(0, 0, self.width, self.height)
			ctx.set_source_rgba(1, 1, 1, self.__flash_phase)
			ctx.fill()
			ctx.save()
			if self.__flash_phase < 0.5:
				self.__flash_phase += 0.5
			else:
				self.__flash_phase = 0.0
		# draw text
		"""ctx.translate(18, 19)
		p_layout = ctx.create_layout()
		p_fdesc = pango.FontDescription("Free Sans Bold 27")
		p_layout.set_font_description(p_fdesc)
		if self.__mail_count > -1:
			mnumstr = str(self.__mail_count)
		else:
			mnumstr = "-"
		p_layout.set_markup(mnumstr)
		ctx.set_source_rgba(0, 0, 0, 0.3)
		ctx.show_layout(p_layout)
		ctx.fill()
		ctx.translate(-1, -1)
		ctx.set_source_rgba(1, 1, 1, 0.8)
		ctx.show_layout(p_layout)
		ctx.fill()"""
		
	def on_draw_shape(self, ctx):
		self.on_draw(ctx)
	
	def on_quit(self):
		# clear timeout
		if self.__timeout:
			gobject.source_remove(self.__timeout)
	

# If the program is run directly or passed as an argument to the python
# interpreter then create a Screenlet instance and show it
if __name__ == "__main__":
	sl = screenlets.create_new_instance('MailCheckScreenlet')
	sl.main()

